(* ::Package:: *)

(* ::Text:: *)
(*(* neuvecv2.m *)*)
(**)
(*(*This package is a variant of pseuvec.m, Version 2 *)*)
(**)
(*(* Created by Rolf Sulanke *)*)
(**)
(*(* pseuvec.m: Started March 18, 1995, last revision September 15, 2003. *)*)
(**)
(*(* pseuvec.m: The package contains vector  operations for the n-dimensional*)
(*pseudoEuclidean space of arbitrary index,  including the Euclidean case,*)
(*it needs euvec.m. For application, examples etc see the notebook *)
(*pseuklid.nb.*)*)
(**)
(*(* neuvec.m: March 8, 2005*)*)
(*(*neuvec.m: In order to have coincidence with the book [2]., the scalar *)
(*product has been changed; the negative terms stand now at the first places.   *)*)
(*(*neuvec.m, October 27, 2009, new Module "indexorder" added.*)*)
(**)
(*(*neuvecv2.m, April 10, 2015, is an adaption to euvecv2.m, tensalgv3.m.Mostly it *)
(*coincides with neuvec.m. *)*)
(**)
(*(* Free software: Feel free to apply, distribute or change the packages and*)
(*notebooks of "DiffGeoCourse". Please, cite the source and notice the*)
(*changes, or new concepts,  which you introduced.*)*)
(**)
(*(* Hint: After importing this package set the numbers*)
(*dim (= dimension) and ind (=index) of the pseudo-Euclidean vector space *)
(*defined as constants within the Global context. They may be changed later in the notebook, if wanted. *)
(*Setting them before importing the package would fix them unchangable in all the following definitons.*)*)
(**)
(*(* WARNINGS. *)
(*1. In certain cases the pseudo-Euclidean orthonormalization procedure orthonorm *)
(*cannot  be carried out. Then instead of the orthonormed matrix appears the *)
(*message  "I cannot orthogonalize the given sequence!"  Sometimes the norming *)
(*is the reason; always try the orthogonalization with the option normed -> False.*)
(* Besides of the parameter dependence the reason can be a numerical badly *)
(*conditioned matrix b (e.g. contains vectors with very small, or very different *)
(*in magnitude, components). In this case one can try applying orthonorm  *)
(*on the renormed matrix renorm[b]. (See the examples in the notebook*)
(* pseuklid.nb.)*)
(*2. In the pseudo-Euclidean case the isotropic vectors are orthogonal to \*)
(*itself. By our definition, they are considered as being normed. Therefore in \*)
(*the resulting orthogonal or orthonormal sequence there may appear linearely \*)
(*dependent isotropic vectors.*)
(*3. In the pseudo-Euclidean vector spaces exist vector sequences which may not be*)
(* orthogonalized preserving its flag structure. In difference to Erhard Schmidt's *)
(*orthogonalization the procedure orthonorm does not preserve the flag structure *)
(*of the given sequence in general.*)*)
(**)
(*(* Changes.*)
(*2015,January 26. The Loading command BeginPackage["pseuvec`",{"vectorcalc`", "euvec`"}] *)
(*replaced by BeginPackage["pseuvec`",{"tensalg`", "euvec`"}]  *)*)
(**)
(*(* Changes for neuvecv2.m*)
(*2015,April 10 26. The Loading command BeginPackage["pseuvec`",{"tensalg`", "euvec`"}}] *)
(*replaced by BeginPackage["neuvecv2`",{"euvecv2`","tensalgv3`"] .*)
(*2017.05.24. ch slightly changed  *)*)


(* ::Section:: *)
(*(*The Usages*)*)


BeginPackage["neuvecv2`",{"euvecv2`","tensalgv3`"}] 
   
(* The usages *)
ide::usage = "ide[d:dim, nd:ind] is a vector of length dim with first  ind entries -1, the other dim - ind:  1; dim  and ind must be defined
first.";

pssp::usage = "pssp[v, w, nd:ind] is the pseudo-Euclidean scalar
product of index ind of the vectors v,w.";

dual::usage = "dual[v, nd:ind] is the dual vector to v with respect
to the scalar product pssp;the index ind must be defined first.";

psgram::usage = "psgram[d:dim, nd:ind] is the Gram matrix of a pseudo-orthonormal basis 
in the pseudo-Euclidean vector space of dimension d and index nd."

pscross::usage = "pscross[veclist]  gives the vector product of a
list veclist of d - 1 vectors (in pseudo-orthogonal coordinates) of dimension d of the pseudo-Euclidean space; 
dim d and ind should be defined first.";

psCross::usage = "psCross[v(1),..., v(d-1)]  gives the 
vector product of d - 1 vectors v(i) (in pseudo-orthogonal coordinates) of dimension d of the
pseudo-Euclidean space;  dim d and ind should be defined first.";

ch::usage = "ch[v] gives 1, if v is spacelike, 0, if v is lightlike, 
and -1, if v is timelike.Warning: Calculating symbolically, for variable v= v[x___]
the singularities v[x]=zero-vector may be neglectected!";

pr::usage = "pr[v,x] is the projection of the vector x onto the 
hypersubspace orthogonal to the vector v with respect to the pseudo-Euclidean
 scalar product pssp; pssp[v,v] != 0 is supposed."   

orthonorm::usage = "orthonorm: If b is a list of vectors in the n-dimensional
 pseudo-Euclidean space of index  k, then orthonorm[b,opts] is an orthogonal basis 
 of span[b]. orthonorm can be applied to finite dimensional \
vector spaces and 
symmetric bilinear forms prod using the option innerprod -> prod; 
the default is prod = pssp."

normalize::usage="normalize[v, prod:pssp] is the proportional vector of the \
same
direction  with prod[v,v] = 0, 1,or -1; in case prod[v,v] = 0 one gets
normalize[v] = v." 

chsort::usage = "chsort[vlist] sorts a list of vectors of the pseudo-Euclideam 
vector space of dimension dim and index ind using the sort function ch; that means: 
first the timelike, then the lightlike (isotropic),  and then the spacelike vectors.
Before sorting the almost zero vectors (see outzero) are deleted.";

psfilter::usage = "psfilter[blist] permutes the vectors of the
list blist; \ the timelike vectors are put at the first places. Besides of
this, their order and the order of the other vectors remain unchanged.
(now obsolete: use chsort."

orthopair::usage = "orthopair[v] constructs a pair of orthonormalized
spacelike vectors starting from  the given spacelike vector v; dim  and ind must
be defined first.";

indexorder::usage = "indexorder[vlist] permutes the row vectors of the 
matrix vlist in such a way that the isotropic vectors go to the first 
places, the timelikes  follow and the spacelikes stand at the end; the 
orders within these groups will not be changed. Before the permutation 
the null vectors will be deleted.";


(* ::Section:: *)
(*(*The Definitions*)*)


Begin["`Private`"]
(* The dimension dim, and the index ind  must be set within the Global Context after loaduing 
the package. *)
dimm := Global`dim; 

indd := Global`ind; 



(* ::Subsubsection:: *)
(*(* Definition of the scalar Product   pssp*)*)


ide[d_:dimm, nd_:indd] := Join[Table[-1, {nd}], Table[1, {d-nd}]];

dual[l_List, nd_:indd] := ide[Length[l], nd] * l
psgram[d_:dimm, nd_:indd] := DiagonalMatrix[ide[d,nd]];

SetAttributes[pssp,Orderless];

pssp[k_List,l_List, nd_:indd]:= dual[k,nd].l;


ch[v_List]:=Module[{nn= Length[v],p,q, v1, v2},
	If[nn<indd, Print["Incorrect argument!"];Return[]];
If[v.v == 0, 
	Print[v, " is the null vector!"]; Return[]]; 
	(*p=Chop[Simplify[Evaluate[pssp[v,v]]]]; *)
	p=Simplify[Evaluate[pssp[v,v]]];
	If[NumberQ[p]&&Im[p]==0, Return[Sign[p]]];	
	v1 = Sum[stb[i,nn]*v[[i]], {i,indd}]; 
	v2 = Sum[stb[j,nn]*v[[j]], {j, indd +1, nn}];
	If[v1 === null[nn] , q = 1; Return[q]];
	If[v2 === null[nn], q = -1; Return[q], q = Sign[p]]; 
	If[ q ===1 || q === -1 || q === 0, Return[q], Print["ch[",v,"] cannot \
    be evaluated!"];Return[]]]

normalize[v_, prod_:pssp]:=Module[{a=Simplify[prod[v,v]], w},
			If[a == 0,w=v,w=v/Sqrt[Abs[a]],w=v/Sqrt[Abs[a]]];
			Return[w]]
			
(* Here are the general cross products in the pseudo-Euclidean case: *)

pscross = Composition[dual, cross]

psCross = Composition[dual, Cross]

(* Orthogonal projection *)
pr[v_, x_] := x - v*pssp[x,v]/pssp[v,v]



(* ::Subsubsection:: *)
(*(*A pseudo-Euclidean orthonormalization procedure. *)*)


Options[orthonorm] = {innerprod ->pssp, normed -> True, print
->False,  neglect-> -10}


orthonorm[b_List,opts___]:=
  Module[{v,w,va,vb,l,ort={},i,j,f,h,sm,aus,prod,nor,pro,bn,follow,iso,
      control},{prod,nor,pro,
        bn}={innerprod,normed,print,neglect}/.{opts}/.Options[orthonorm];
    f[h_]:=smoothing[h,10^bn];
    sm[aus_]:=MapAll[f,aus];
    w=outzero[sm[b]];
    Label[follow];
    l=Length[w];
    If[l==0,Goto[control]];
    For[i=1,i<=l,i++,
      If[Not[Simplify[sm[prod[w[[i]],w[[i]]]]]==0],
        AppendTo[ort,w[[i]]];
        If[pro==True,Print[w[[i]]]];
        v=
          Simplify[
            sm[Table[
                w[[j]]-w[[i]]*prod[w[[i]],w[[j]]]/prod[w[[i]],w[[i]]],{j,
                  l}]]];
        w=outzero[v]; Goto[follow],Null,AppendTo[ort,w[[i]]];
        If[pro==True,Print[w[[i]]]];
        v=
          Simplify[
            sm[Table[
                w[[j]]-w[[i]]*prod[w[[i]],w[[j]]]/prod[w[[i]],w[[i]]],{j,
                  l}]]];
        w=outzero[v]; Goto[follow]]];
    Label[iso];l=Length[w];
    If[l==0,Goto[control]];
    v=w;
    If[l==1,AppendTo[ort,v[[1]]]; If[pro==True,Print[v[[1]]]];
      Goto[control]];
    For[i=1,i<=l,i++,
      For[j=i+1,j<=l,j++,
        If[Not[Simplify[sm[prod[v[[i]],v[[j]]]]]==0],va=v[[j]]-v[[i]];
          vb=v[[j]]+v[[i]]; 				v[[i]]=va;v[[j]]=vb;
          w=outzero[Simplify[sm[v]]];Goto[follow]]]];
          v=basis[v]; 			
          ort=Join[ort,v];
     Label[control];
    l=Length[ort];
    If[l==0,Return[ort]];
    If[nor==True,
      ort=sm[Table[Simplify[normalize[ort[[i]],prod]],{i,Length[ort]}]]];
    For[i=1,i<=l-1,i++,
      For[j=i+1,j<=l,j++,
        If[Not[Simplify[sm[prod[ort[[i]],ort[[j]]]]]==0],
          Return[Print["I cannot orthogonalize the given sequence!"]],Null,
          Return[Print["I cannot orthogonalize the given sequence!"]]]]]; 	
    Return[ort]]


indexorder[v_List]:=Module[{m , n = Length[v[[1]]],w,ww, ze =0,begin},
If[MatrixQ[v] == False ||  n!= dim, Print["Incorrect argument!"];Return[]];
w = outzero[v];m= Length[w];
Label[begin];
If[ze >= m^2, Return[w]];
ww = w;
For[ j =2,j<=m, j++,  If[ch[w[[j]] ]==0 && ch[w[[j-1]]] !=0, ww[[j-1]] =w[[j]];ww[[j]] =w[[j-1]]];
 If[ch[w[[j]] ]<0 && ch[w[[j-1]]] >0, ww[[j-1]] =w[[j]];ww[[j]] =w[[j-1]]];w=ww; ze = ze+1];
Goto[begin];]


(* ::Subsubsection:: *)
(*(*psfilter, chsort: rearranging the sequence: timelike vectors ahead! *)*)


psfilter[b_List] := Module[{i, lb = Length[b], bn= {},qn = {}},
		For[i=1,i<=lb,i++,
			If[Length[b[[i]] ] != dimm, 
        Return[ Print["Incorrect argument: ", b[[i]],"!!!"]]]];
		For[i=1, i<=lb, i++,
			If[ch[b[[i]] ]== -1, AppendTo[bn, b[[i]]],AppendTo[qn,b[[i]]],
        AppendTo[qn,b[[i]]]]];
				Return[Join[bn,qn]]]

chsort[b_List] := Sort[outzero[b],ch[#1]<=ch[#2]&];


(* ::Subsubsection:: *)
(*orthopair*)


(* ::Text:: *)
(*(* Now we generate a pair of orthonormal spacelike vectors starting from a *)
(*given spacelike one; *)
(*that we need for the geometry of circles on the 3-sphere. *)*)


orthopair[h_List] := 
Module[{a , kk , m1, b = dimm - indd, j, i,mk},
ml = null[dimm];	
a = Simplify[pssp[h,h]];
If[a <= 0, Print["Vector ",h," is not spacelike!"];Return[], Null,
Print["Maybe, Vector ",h," is not spacelike?"];Return[]]; 
If[ b < 2, Print["Orthogonal spacelike pairs don't exist!"];Return[]]; 	 
kk = h/Sqrt[a]; 	
	 	For[i = indd+1, i <= dimm, i++, 			
	If[kk[[i]] == 0, m1 = stb[i]; Return[{kk, m1}]]];   
ml[[dimm -1]] = -kk[[dimm]];
ml[[dimm]] = kk[[dimm-1]]; 
		mk = ml/Sqrt[pssp[ml, ml]];  				
Return[{kk, mk}]]/; Length[h] == dimm

End[]	
 Protect["`*"] ;

EndPackage[]

